signature PP = sig

(* Pretty-printing stream abstraction.  This formulation is
   similar to Derek Oppen's pretty-printing streams, but is actually
   based on hand-written notes of Greg Nelson [CGN 8, 9, and 23, 
   Nov-Dec 1983] *)

structure Wr: WR

type Pp

(* ---------------------------------------------------------------------- *)
(* Creation and coercion *)

(* Build a new Pp whose underlying target for formatted lines is 
   the given Wr.T *)
val pp_from_wr: Wr.Wr -> Pp

(* Build a new Wr.T whose target is the given Pp.  This allows a
   "connected pair" of a writer and a prettyprinter to be defined by
   building the prettyprinter first and then calling wr_from_pp to
   build the writer.  Think of this as a poor man's subtyping: 
   wr_from_pp is the canonical coercion function from Pp to Wr.T. *)
val wr_from_pp: Pp -> Wr.Wr;


(* ---------------------------------------------------------------------- *)
(* Output and formatting *)

(* Send a string of characters to a Pp.  Newline characters in the string
   have the effect of inserting a breakpoint and forcing it to be chosen. *)
val ppwrite : Pp -> string -> unit

(* Mark the beginning of a group.  If a group fits together on a line,
   it will be formatted together.  Otherwise ALL of the outermost united 
   breakpoints in the group are broken. *)
val ppsetb: Pp -> unit

(* Mark the end of a group. *)
val ppendb: Pp -> unit

(* Insert a breakpoint in the current group.  The boolean flag controls
   whether this breakpoint is "united" with the other breakpoints in
   the group: when any breakpoint is broken, all united breakpoints in
   its group are also broken.  Ununited breakpoints are only broken 
   when the text following them would overflow a line; they can be used
   to fill lines with words, for example.  
   
   If this breakpoint is chosen, the indentation of the new line is
   found by adding the integer parameter to the indentation in effect at 
   the start of the group. *)
val ppbreak: Pp -> bool -> int -> unit

(* Insert a breakpoint with an explicit string.  If this breakpoint is
   chosen, the string parameter is inserted in the output stream as if 
   it had been provided as an argument to ppwrite at this point.  
   Newline characters in the string can be used to actually break the 
   line, if desired. *)
val ppexpbreak: Pp -> bool -> string -> unit


(* Define the width of the page. *)
val ppsetmargin: Pp -> int -> unit


(* ---------------------------------------------------------------------- *)

(* Internal debugging flag *)
val DEBUG: bool ref

end
