\begin{exercise}
Consider the type:
> Ex = Some(A) {|e:A, f:A->Int->Int|};
A typical value of this type is
> p = < Int,
>       {e = 5, 
>           f = fun(x:Int) fun(y:Int) x}
>     >: Ex;
Find two more expressions with this type.
\end{exercise}

\ifsolution
\noindent For example:
> <Int, {e = 5, f = fun(x:Int) fun(y:Int) plus x y}> : Ex;
> <Bool, {e = true, f = fun(x:Bool) fun(y:Int) y}> : Ex;
\fi

\begin{exercise}
\ifsolution\else
Given an element of <<Ex>>, the following function opens it and
applies the <<f>> component to the <<e>> component:
> use_it = fun(p:Ex)
>            open p as <A,r> in
>              r.f r.e 2
>            end;
> use_it p;
\fi
Write a function that takes an element of <<Ex>> and uses it to build
an element of the type <<Some(A) {|e:A, f:A->Int|}>>.
\end{exercise}

\ifsolution
> fun(p:Ex)
>   open p as <A,r> in
>     < A,
>       {e = r.e,
>        f = fun(a:A) r.f a 5}
>     >: Some(A) {|e:A, f:A->Int|}
>   end;
\fi

\begin{exercise}\ifsolution\else\label{ex:packing}\fi
How many different values for <<T>> can you find such that the
expression 
\begin{verbatim}
    <Int, {a=5,b=7}> : Some(A)T 
\end{verbatim}
is well typed?
\end{exercise}

\ifsolution
\noindent Here are all the ones I can think of:
> <Int, {a=5,b=7}> : Some(A){|a:A,b:A|};
> <Int, {a=5,b=7}> : Some(A){|a:Int,b:A|};
> <Int, {a=5,b=7}> : Some(A){|a:A,b:Int|};
> <Int, {a=5,b=7}> : Some(A){|a:Int,b:Int|};
\fi

\begin{exercise}\label{ex:packing}
Is there any pair of values for <<T>> and <<T'>> where <<T'>> is {\em
not} <<Int>> and the expression
\begin{verbatim}
    <T', {a=5,b=7}> : Some(A)T 
\end{verbatim}
is well typed?
\end{exercise}

\ifsolution
\noindent Yes, infinitely many.  For example:
> <Bool, {a=5,b=7}> : Some(A){|a:Int,b:Int|};
\fi

\begin{exercise}
Is the expression
\begin{verbatim}
    let p = <Int, {a=5}> : Some(A) {|a:A|} 
    in open p as <R,r>
       in 
         {c = r.a, d = true}
       end
    end
\end{verbatim}
well typed?  
\end{exercise}

\ifsolution
>> reset haltonerror;
> let p = <Int, {a=5}> : Some(A) {a:A} 
> in open p as <A,r>
>     in 
>       {c = r.a, d = true}
>     end
>  end;
>> set haltonerror;
\fi

\begin{exercise}
Recall the definition of the <<Point>> type:
> Point = Some(R) 
>            {| state: R, methods: {| setX: R->Int->R, getX:R->Int |}|};
\ifsolution\else
Since we will be writing many such type expressions, it will be
convenient to separate them into a common part that describes their
outer structure (the existential type and the outer record) and a
specific part that lists the methods and their types (the inner
record).  These parts will then be combined to produce specific object
types such as <<Point>>.
\fi
Define the type operator <<Object:(*->*)->*>> and the type
<<PointM:*->*>> such that
\begin{verbatim}
    Point = Object(PointM)
\end{verbatim}
\end{exercise}

\ifsolution
> Object = Fun(M:*->*) Some(R) {|state: R, methods: M(R)|};
> PointM = Fun(R) {|setX: R->Int->R, getX: R->Int|};
> check Point ~ Object(PointM);
\fi

\begin{exercise}\ifsolution\else\label{ex:slidepoints}\fi
Use the <<Object>> operator and a new operator <<SlidePointM>> to
give an extension of the type <<Point>> with a <<slide>> method of
the same type as <<setX>>, 
\end{exercise}

\ifsolution
> SlidePointM = Fun(R) {|setX: R->Int->R, getX: R->Int, slide: R->Int->R|};
> SlidePoint = Object(SlidePointM);
\fi

\begin{exercise}
\ifsolution\else
In class we built a typical element of <<Point>> like this:
> p = < Int,
>       {state = 5,
>        methods = {setX = fun(x:Int) fun(i:Int) i,
>                   getX = fun(x:Int) x}}
>     >: Point ;
\fi
Show how to build a typical element of <<Object(SlidePointM)>>.
\end{exercise}

\ifsolution
> mp = < Int,
>        {state = 5,
>         methods = {setX = fun(x:Int) fun(i:Int) i,
>                    slide = fun(x:Int) fun(i:Int) plus x i,
>                    getX = fun(x:Int) x}}
>     >: Object(SlidePointM) ;
\fi

\begin{exercise}
\ifsolution\else
Recall the definition of the <<Point'setX>> function, which ``sends
the message <<setX>>'' to elements of <<Object(PointM)>>:
> Point'setX =
>   fun(p:Point)
>     fun(i:Int)
>       open p as <R,r> in
>         < R, 
>           {state = r.methods.setX r.state i,
>            methods = r.methods}
>         >: Point 
>       end;
The definition of <<SlidePoint'setX>> is identical except for the type
of the first parameter.
\fi
Define <<SlidePoint'slide>> and <<SlidePoint'getX>>.
\end{exercise}

\ifsolution
> SlidePoint'slide =
>   fun(p:Object(SlidePointM))
>     fun(i:Int)
>       open p as <R,r> in
>         < R, 
>           {state = r.methods.slide r.state i,
>            methods = r.methods}
>         >: Object(SlidePointM)
>        end;
> SlidePoint'getX =
>   fun(p:Object(SlidePointM))
>       open p as <R,r> in
>         r.methods.getX r.state
>       end;
\fi

\begin{exercise} (Optional) \ifsolution\else\label{ex:simpdele}\fi
\ifsolution\else
Here is an interface specification for colored points:
> CPointM = Fun(R) {| setX: R->Int->R, getX: R->Int,
>                     setC: R->Color->R, getC: R->Color |};
\fi
Write a function that takes an element of <<Object(PointM)>> and uses
it to build an element of <<Object(CPointM)>>.
\end{exercise}

\ifsolution
\noindent Here are two possibilities:
>> !besilent;
> Point'getX =
>   fun(p:Object(PointM))
>       open p as <R,r> in
>         r.methods.getX r.state
>       end;
>> ! beloud;
> cpoint_from_point =
>   fun(p:Object(PointM))
>     < {| pstate:Object(PointM), cstate:Color |}, 
>       {state = {pstate = p,
>                 cstate = red},
>        methods = 
>           {setX = fun(s:{|pstate:Object(PointM),cstate:Color|}) fun(i:Int)
>                     let new_pstate = Point'setX s.pstate i
>                     in {pstate=new_pstate, cstate=s.cstate}
>                     end,
>            getX = fun(s:{|pstate:Object(PointM),cstate:Color|}) 
>                     Point'getX s.pstate,
>            setC = fun(s:{|pstate:Object(PointM),cstate:Color|}) fun(c:Color)
>                     {pstate=s.pstate, cstate=c},
>            getC = fun(s:{|pstate:Object(PointM),cstate:Color|}) 
>                     s.cstate}}
>     >: Object(CPointM);

> cpoint_from_point' =
>   fun(p:Object(PointM))
>     open p as <R,r> in
>       < {|pstate:R, cstate:Color|}, 
>         {state = {pstate = r.state,
>                   cstate = red},
>          methods = 
>             {setX = fun(s:{|pstate:R,cstate:Color|}) fun(i:Int)
>                       let new_pstate = r.methods.setX s.pstate i
>                       in {pstate=new_pstate, cstate=s.cstate}
>                       end,
>              getX = fun(s:{|pstate:R,cstate:Color|}) 
>                       r.methods.getX s.pstate,
>              setC = fun(s:{|pstate:R,cstate:Color|}) fun(c:Color)
>                       {pstate=s.pstate, cstate=c},
>              getC = fun(s:{|pstate:R,cstate:Color|}) 
>                       s.cstate}}
>      >: Object(CPointM)
>     end;
\fi


