\begin{exercise}
The predefined operator <<plus>> forms the sum of two integer
arguments:
> plus 2 3;
Use <<plus>> to compute the sum of 7, 8, 9, and 10.
\end{exercise}

\ifsolution
\noindent
One possibility:
> plus (plus 7 8) (plus 9 10);
\fi

\begin{exercise}
The definition
> double = fun(x:Int) plus x x;
defines a new doubling function:
> double 5;
Use double to define a function that multiplies its argument by four.
\end{exercise}

\ifsolution
> quad = fun(x:Int) double x;
\fi

\begin{exercise}
Records are built using curly braces:
> r = {a=5, b=7};
The field selection operator is dot (``<<.>>''):
> plus r.a r.b;
Write a function that takes as argument a record with two integer
fields named <<a>> and <<b>> and returns a record with two fields
containing the sum and product of the fields of the argument.
\end{exercise}

\ifsolution
> f = fun(r:{a:Int,b:Int}) 
>       {sum = plus r.a r.b,
>        product = times r.a r.b};
> f r;
\fi

\begin{exercise}
The polymorphic identity function
> id = fun(A) fun(x:A) x;
can be used to obtain the integer identity function by instantiating
it at the type <<Int>>:
> id Int;
> id Int 3;
Write a function that takes an argument of type <<All(A) A->A>> and
(by instantiating it appropriately) applies it to itself.
\end{exercise}

\ifsolution
> delta = fun(f:All(A) A->A) f (All(A) A->A) f;
> delta id;
\fi

\begin{exercise}
\ifsolution\else
The <<fomega>> system includes a facility for simple type
abbreviations.  For example, the declaration
> IntFun = Int->Int;
defines a type identifier <<IntFun>> that behaves identically in all
contexts with the longer expression <<Int->Int>>:
> g = fun(f:IntFun) fun(x:Int) f (f x);
Note that the type of <<g>> is equivalent to <<IntFun->IntFun>>, but
this fact is not noticed by the type synthesis algorithm, which 
{\em expands} abbreviations as necessary but never tries to contract
them.  In situations where the abbreviated form is preferable, you can
explicitly annotate parts of your program with types that you
believe they possess; these assertions are verified by the
typechecker.  For example:
> g = fun(f:IntFun) 
>       (fun(x:Int) f (f x)) 
>       :<IntFun>;
\fi
Define an abbreviation <<Id>> for the type of the polymorphic identity
function above.  Use it to write a version of the polymorphic
self-application function whose type is printed as <<Id->Id>>.  Define
another version whose type is printed as <<Id->(All(A)A->A)>> (there
are at least two ways to do this).
\end{exercise}

\ifsolution
> Id = All(A) A->A;
> delta = fun(x:Id) x Id x;
> delta' = fun(x:Id) x (All(A)A->A) x;
> delta'' = fun(x:Id) (x Id x) :<All(A)A->A>;
\fi

\begin{exercise}
\ifsolution\else
One of the interesting properties of the polymorphic lambda-calculus
is that it can be used to encode many basic datatypes, such as
numbers, booleans, etc.  For example, the type of the ``Church
booleans''
(c.f.~\cite[etc.]{BOHM85,Church41:CalcOfLambdaConv,PIERCE89}) is:
> CBool = All(A) A->A->A;
This type has two normal-form inhabitants:
> ctrue = fun(A) fun(t:A) fun(f:A) t;
> cfalse = fun(A) fun(t:A) fun(f:A) f;
Then, for example, the <<not>> function on booleans can be written
> cnot = fun(b:CBool) b CBool cfalse ctrue;
\fi
Write a function <<cand : CBool->CBool->CBool>> that computes the
logical ``and'' of its arguments.  Write a three-argument logical ``and''
function. 
\end{exercise}

\ifsolution
> cand = fun(a:CBool) fun(b:CBool)
>          a CBool b cfalse
>          :<CBool>;
> cand ctrue cfalse;
> cand3 = fun(a:CBool) fun(b:CBool) fun(c:CBool)
>          a CBool (b CBool c cfalse) cfalse
>          :<CBool>;
> cand3 ctrue ctrue ctrue;
\fi

\begin{exercise}
\ifsolution\else
The <<fomega>> implementation also provides a built-in type <<Bool>>
with the predefined constants <<true>> and <<false>> and a primitive
selector <<primIf : All(A) A->A->A>>.  The parser allows the special
syntax
\begin{verbatim}
    if T b then e1 else e2 end
\end{verbatim}
as an abbreviation for
\begin{verbatim}
    primIf T b e1 e2
\end{verbatim}
\fi
Use the <<if...then...else...end>> syntax to define an <<and>>
function for pairs of booleans.
\end{exercise}

\ifsolution
> and = fun(a:Bool) fun(b:Bool)
>         if Bool a 
>            then b
>            else false
>         end;
> and true false;
\fi

\begin{exercise}
\ifsolution\else
The type of pairs of integers can be defined following the model of the
Church booleans:
> IntPair = All(A) (Int->Int->A) -> A;
> pair = fun(i:Int) fun(j:Int)
>          (fun(A) fun(f:Int->Int->A) 
>             f i j)
>          :<IntPair>;
> p = pair 3 5;
> fst = fun(p:IntPair) p Int (fun(i:Int) fun(j:Int) i);
\fi
Define the <<snd>> function on pairs.
\end{exercise}

\ifsolution
> snd = fun(p:IntPair) p Int (fun(i:Int) fun(j:Int) j);
> times (fst p) (snd p);
\fi

\begin{exercise}
\ifsolution\else
The pervasive environment includes a polymorphic fixed point operator
> rec;
which, given a type <<A>> and a function <<f : A->A>>, computes the
``least'' value <<x>> such that <<f x = x>>.  \fi
Use <<rec>> to write a
function that computes the factorial of its argument.  \ifsolution\else
(You will 
need some of the other pervasive arithmetic functions listed in
the {\em F-Omega-Sub User's Manual}.  You may also want to put your
definition in a file rather than typing it directly to the
interpreter.  The <<use>> command can be used to load the file into a
session with the <<fomega>> system, or the file may be processed by
typing  <<fomega filename>> from the unix command line.)
\fi
\end{exercise}

\ifsolution
> fact = 
>   rec (Int->Int)
>      fun(f:Int->Int)
>        fun(n:Int)
>          if Int eqInt n 0
>            then 1
>            else times n (f (minus n 1))
>          end;
> fact 6;
\fi

\begin{exercise}
\ifsolution\else
The <<fomega>> system also supports higher-order polymorphism.  For
example,
> Pair = Fun(T:*) All(A) (T->T->A) -> A;
> pair = fun(T)
>          fun(i:T) fun(j:T)
>            (fun(A) fun(f:T->T->A) 
>               f i j)
>            :<Pair(T)>;
> p = pair Int 3 5;
> fst = fun(T) fun(p: Pair T) p T (fun(i:T) fun(j:T) i);
> snd = fun(T) fun(p: Pair T) p T (fun(i:T) fun(j:T) j);
\fi
Extend the definition of polymorphic pairs to the case where the first
and second projections may have different types.
\end{exercise}

\ifsolution
> Pair = Fun(S:*) Fun(T:*) All(A) (S->T->A) -> A;
> pair = fun(S) fun(T)
>          fun(i:S) fun(j:T)
>            (fun(A) fun(f:S->T->A) 
>               f i j)
>            :<Pair S T>;
> p = pair Bool Int true 5;
> fst = fun(S) fun(T) fun(p: Pair S T) p S (fun(i:S) fun(j:T) i);
> snd = fun(S) fun(T) fun(p: Pair S T) p T (fun(i:S) fun(j:T) j);
\fi
