functor Leq(
	  structure Typ: TYP
	  structure TypHelper: TYPHELPER
	  structure TEnv: TENV
	  structure Globals: GLOBALS
	  sharing Typ.Globals = Globals
	      and TypHelper.TEnv = TEnv
	      and TEnv.Typ = Typ
	      and Globals.Id = Typ.Id
	  ) : LEQ = struct

structure Typ = Typ
structure Globals = Globals
open Globals
open TypHelper
open TEnv
open Typ

datatype lhsqueue = 
	    ARROW_LHS of Typ.T
	  | ALL_LHS   of Id.T * Typ.T
	  | OBJ_LHS   of Id.T

datatype rhs_flag = EXPAND | FIX

val DEBUG = createflag ["leq","Leq"] false

(* $$$ very inefficient! *)
fun idset_eq L1 L2 = 
    (forall (fn l1 => memq Id.== L2 l1) L1)
    andalso
    (forall (fn l2 => memq Id.== L1 l2) L2)

fun prt_idset L = 
    (write "{";
     mapunit_tuple (fn i=> write (Id.tostr i)) (fn()=> write",") L;
     write "}")

fun describe_rest te [] t Lt  = 
      (endb ();
       write "]";
       break true ~1;
       write " -> ";
       prt te t;
       write " ! ";
       prt_idset Lt)
  | describe_rest te [ARROW_LHS(t1)] t2 Lt = 
      (prt te t1;
       describe_rest te [] t2 Lt)
  | describe_rest te ((ARROW_LHS(t1))::X2) t2 Lt = 
      (prt te t1;
       write ",";
       break true 0;
       describe_rest te X2 t2 Lt)
  | describe_rest te [ALL_LHS(v,t1)] t2 Lt  =  let
      val te' = (push_bound te v t1)
      in
      (write (Id.tostr (lookup_name te' 0));
       write "<";
       prt te t1;
       describe_rest te' [] t2 Lt)
      end
  | describe_rest te ((ALL_LHS(v,t1))::X2) t2 Lt  = let
      val te' = (push_bound te v t1)
      in
      (write (Id.tostr (lookup_name te' 0));
       write "<";
       prt te t1;
       write ",";
       break true 0;
       describe_rest te' X2 t2 Lt)
      end
  | describe_rest te [OBJ_LHS(i)] t2 Lt = 
      (write (Id.tostr i);
       describe_rest te [] t2 Lt)
  | describe_rest te (OBJ_LHS(i)::X2) t2 Lt = 
      (write (Id.tostr i);
       write ",";
       break true 0;
       describe_rest te X2 t2 Lt)

fun describe_problem te s Ls X t Lt =
  (setb ();
   prt te s;
   write " ! ";
   prt_idset Ls;
   break true ~4;
   write " < ";
   write "[";
   setb ();
   describe_rest te X t Lt;
   endb ())

fun rebuild [] t = t
  | rebuild ((ARROW_LHS(t1))::X2) t = arrow t1 (rebuild X2 t)
  | rebuild ((ALL_LHS(v,t1))::X2) t = all v t1 (rebuild X2 t)
  | rebuild ((OBJ_LHS(i))::X2) t = obj i (rebuild X2 t)

fun bindings_in [] = 0
  | bindings_in (ARROW_LHS(_)::tl) = bindings_in tl
  | bindings_in (ALL_LHS(_)::tl) = 1 + (bindings_in tl)
  | bindings_in (OBJ_LHS(_)::tl) = (bindings_in tl)

fun leqqU' te s Ls X t Lt =
  case explode(t ) of
        EXPTVAR vt => let 
      val bx = bindings_in X
      in if vt < bx
	 then leqqL te s Ls X t Lt
	 else lookup_and_relocate te (vt - bx) 
	       (fn _ => leqqL te s Ls X t Lt)
	       (fn _ => raise WrongKindOfId(te, vt - bx,"tvar or tabbrev"))
	       (fn ab => leqqU te s Ls X ab Lt)
      end
      | EXPARROW (t1,t2) => 
      leqqU te s Ls (X@[ARROW_LHS(t1)]) t2 []
      | EXPALL (i,t1,t2) => 
      leqqU te s Ls (X@[ALL_LHS(i,t1)]) t2 []
      | EXPSOME (i,t1,t2) => 
      leqqL te s Ls X t Lt
      | EXPMEET ts => 
      forall (fn t => leqqU te s Ls X t Lt) ts
      | EXPOBJ (i,t) => 
      if memq Id.== Lt i
	  then true
          else leqqU te s Ls (X@[OBJ_LHS(i)]) t []
      | EXPMINUS (t,ids) => 
      leqqU te s Ls X t (Lt @ ids)

and leqqL' te s Ls X t Lt =
  case explode(s) of
      EXPTVAR vs => 
      ((null X) 
       andalso 
       case explode(t) of
             EXPTVAR vt => vs = vt
           | EXPARROW _ => false
           | EXPALL _ => false
           | EXPSOME _ => false
           | EXPMEET _ => false
           | EXPOBJ _ => false
           | EXPMINUS _ => false
       andalso
       (idset_eq Ls Lt))
      orelse
      (lookup_and_relocate te vs 
         (fn bnd => (leqqL te bnd Ls X t Lt))
         (fn _ => raise WrongKindOfId(te,vs,"tvar or tabbrev"))
         (fn ab => (leqqL te ab Ls X t Lt)))
    | EXPARROW (s1,s2) => 
      (case X of
        (ARROW_LHS(t1)::X') =>
            (leq te t1 s1) andalso (leqqL te s2 [] X' t Lt)
      | _ => 
            false)
    | EXPALL (i,s1,s2) => 
      (case X of
        (ALL_LHS(i',t1)::X') =>
            (leq te t1 s1) andalso (leqqL (push_bound te i' t1) s2 [] X' t Lt)
      |	_ => 
            false)
    | EXPSOME (i,s1,s2) => 
      (case explode(t ) of
             EXPTVAR _ => false
           | EXPARROW _ => false
           | EXPALL _ => false
           | EXPSOME (n,t1,t2) => (null X) 
          		  andalso (leq te s1 t1)
          		  andalso (leq (push_bound te n s1) s2 t2)
           | EXPMEET _ => false
           | EXPOBJ _ => false
           | EXPMINUS _ => false )
    | EXPMEET ss => 
      forsome (fn s => leqqL te s Ls X t Lt) ss
    | EXPOBJ (ns,s1) => 
      (case X of
        (OBJ_LHS(nt)::X') =>
            (Id.== ns nt) 
	    andalso (not (memq Id.== Ls ns))
	    andalso (leqqL te s1 [] X' t Lt)
      |	_ => 
            false)
    | EXPMINUS (s1,ids) => 
      leqqU te s1 (Ls @ ids) X t Lt

and leqqU te s Ls X t Lt =
  wrap DEBUG "leqqU"
    (fn () => 
      leqqU' te s Ls X t Lt)
    (fn () => describe_problem te s Ls X t Lt)
    (fn b => write (if b then "Yes" else "No"))
   
and leqqL te s Ls X t Lt =
  wrap DEBUG "leqqL"
    (fn () => 
      leqqL' te s Ls X t Lt)
    (fn () => describe_problem te s Ls X t Lt)
    (fn b => write (if b then "Yes" else "No"))
   
and leq te s t = 
  wrap DEBUG "leq"
    (fn () => 
      leqqU te s [] [] t [])
    (fn () => 
       (setb (); prt te s; break true ~4; write " < "; prt te t; write "? ";
	endb ()))
    (fn b => write (if b then "Yes" else "No"))

fun eqv te s t = (leq te s t) andalso (leq te t s)

end
