functor TrmHelper(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    sharing Typ.Globals = Globals
	        and TypHelper.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
		and Trm.TypHelper = TypHelper
	        and Globals.Id = Typ.Id
	        and Typ.Id = TEnv.Id
	        and TEnv.Id = TypHelper.Id
	    ) : TRMHELPER
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TypHelper = TypHelper
structure TEnv = TEnv
structure Trm = Trm

open Globals
open Trm
open TypHelper
open TEnv
open Typ
open Pp

datatype pretrm = 
	    PREVAR of Id.T
	  | PREABS of Id.T * TypHelper.pretyp * pretrm
	  | PREAPP of pretrm * pretrm
	  | PRETABS of Id.T * TypHelper.pretyp * pretrm
	  | PRETAPP of pretrm * TypHelper.pretyp
	  | PREFOR of Id.T * (TypHelper.pretyp list) * pretrm
	  | PREQUERY of string * pretrm 
	  | PREPROJ of pretrm * Id.T 
	  | PREWITH of pretrm * Id.T * pretrm
	  | PREOPEN of Id.T * Id.T * pretrm * pretrm
	  | PREPACK of pretrm * TypHelper.pretyp * TypHelper.pretyp
	  | PRENOTHING

fun debruijnify te (PREVAR i) =
      VAR((), index te i)
  | debruijnify te (PREABS(i,ptyp,ptrm)) =
      ABS({name=i}, TypHelper.debruijnify te ptyp, 
      		    debruijnify (push_binding te i NS) ptrm)
  | debruijnify te (PREAPP(ptrm1,ptrm2)) =
      APP((), debruijnify te ptrm1, debruijnify te ptrm2)
  | debruijnify te (PRETABS(i,ptyp,ptrm)) =
      TABS({name=i}, TypHelper.debruijnify te ptyp, 
      		     debruijnify (push_bound te i NS) ptrm)
  | debruijnify te (PRETAPP(ptrm,ptyp)) =
      TAPP((), debruijnify te ptrm, TypHelper.debruijnify te ptyp)
  | debruijnify te (PREFOR(i,ptyps,ptrm)) =
      FOR({name=i}, map (fn pt => TypHelper.debruijnify te pt) ptyps, 
      		     debruijnify (push_bound te i NS) ptrm)
  | debruijnify te (PREQUERY(tag,ptrm)) = 
      QUERY(tag, debruijnify te ptrm) 
  | debruijnify te (PREPROJ(ptrm,n)) = 
      PROJ(debruijnify te ptrm, n) 
  | debruijnify te (PREWITH(ptrm1,i,ptrm2)) = 
      WITH(debruijnify te ptrm1, i, debruijnify te ptrm2) 
  | debruijnify te (PREOPEN(a,x,e1,e2)) = 
      OPEN(a, x, debruijnify te e1, 
      	   debruijnify (push_binding (push_bound te a NS) x NS) e2)
  | debruijnify te (PREPACK(e,t,h)) = 
      PACK(debruijnify te e, TypHelper.debruijnify te t, TypHelper.debruijnify te h) 
  | debruijnify te (PRENOTHING) =
      NOTHING

fun prt te trm =
  let fun p te (VAR(_,i)) =
              write (Id.tostr (lookup_name te i))
        | p te (ABS({name=i},t,body)) =
              (write "(\\";
               write (Id.tostr i);
               write ":";
               TypHelper.prt te t;
               write ". ";
               p (push_binding te i t) body;
               write ")")
        | p te (APP(_,trm1,trm2)) =
              (write "(";
               p te trm1;
               write " ";
               p te trm2;
               write ")")
        | p te (TABS({name=i},t,body)) =
              (write "(\\\\";
               write (Id.tostr i);
               write "<";
               TypHelper.prt te t;
               write ". ";
               p (push_bound te i t) body;
               write ")")
        | p te (TAPP(_,trm1,t)) =
              (write "(";
               p te trm1;
               write " [";
               TypHelper.prt te t;
               write "])")
        | p te (FOR({name=i},ts,body)) =
              (write "(for ";
               write (Id.tostr i);
               write " in ";
               mapunit_tuple (fn t => TypHelper.prt te t) (fn () => write ",") ts;
               write ". ";
               p (push_abbrev te i NS) body;
               write ")")
        | p te (PROJ(body,n)) = 
              (write "("; 
               p te body; 
               write "."; 
               write (Id.tostr n); 
               write ")") 
        | p te (OPEN(a,x,e1,e2)) = 
              (write "(open "; 
               p te e1; 
               write " as [";
               write (Id.tostr a);
               write ",";
               write (Id.tostr x);
               write "] in ";
               p te e2; 
               write " end)") 
        | p te (PACK(e,t,h)) = 
              (write "(pack "; 
               p te e; 
               write " as "; 
               TypHelper.prt te t;
               write " hiding "; 
               TypHelper.prt te h;
               write " end)") 
        | p _ _ =
              (write "<trm>")
  in p te trm
  end

end
