functor Object(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    structure TrmHelper: TRMHELPER
	    structure Eval: EVAL
	    structure Synth: SYNTH
	    sharing Typ.Globals = Globals
	        and TypHelper.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
		and Trm.TypHelper = TypHelper
	        and Globals.Id = Typ.Id
	        and Typ.Id = TEnv.Id
	        and TEnv.Id = TypHelper.Id
	        and TrmHelper.Trm = Trm
	        and Typ.Globals = Globals
	        and TEnv = Eval.TEnv
	        and TEnv.Globals = Globals
	        and Synth.Typ = Typ
	        and Synth.TEnv = TEnv
	        and Synth.Trm = Trm
	        and TrmHelper.Id = Globals.Id
	        and TrmHelper.TypHelper = TypHelper
	        and Eval.Id = Globals.Id
	        and Eval.Trm = Trm
	    ) : OBJECT
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TypHelper = TypHelper
structure TEnv = TEnv
structure Trm = Trm
structure TrmHelper = TrmHelper
structure Eval = Eval

open Eval
open TEnv
open TrmHelper
open Trm
open TypHelper
open Typ

exception BadObjDecl of string

val DEBUG = createflag ["object"] false
val MUCHODEBUG = createflag ["OBJECT"] false

(* NB: same name as in toplevel.sml *)
val eval = createflag ["eval"] true

(* Should be able to define LIFT and LOWER using only substitution. *)
fun lift te i t Object = let
  fun p (TVAR(i')) = 
        if (i = i') 
           then appt (tvar(Object)) (tvar(i'+2))
           else tvar(i')
    | p (ABST(n,k1,t2)) = 
          t
    | p (APPT(t1,t2)) = 
          t
    | p (ARROW(t1,t2)) = 
          arrow (lower te i t1 Object) (lift te i t2 Object)
    | p (ALL(x,t1,t2)) = 
          all x t1 (lift (push_bound te x dummytype) (i+1) t2 (Object+1))
    | p (SOM(x,t1,t2)) = 
          some x t1 (lift (push_bound te x dummytype) (i+1) t2 (Object+1))
    | p (TOP(k)) =
          top(k)
    | p (RCD(fields)) = 
          rcd (map (fn (l,t) => (l, lift te i t Object)) fields)
  in wrap MUCHODEBUG "lift"
    (fn () => p (whn te t))
    (fn () => (write (makestring i); write " in "; prt te t))
    (fn t' => prt te t')
  end

and lower te i t Object = let
  fun p (TVAR(i')) = 
        if (i = i') 
           then raise BadObjDecl ("representation type appears negatively")
           else tvar(i')
    | p (ABST(n,k1,t2)) = 
          t
    | p (APPT(t1,t2)) = 
          t
    | p (ARROW(t1,t2)) = 
          arrow (lift te i t1 Object) (lower te i t2 Object)
    | p (ALL(x,t1,t2)) = 
          all x t1 (lower (push_bound te x dummytype) (i+1) t2 (Object+1))
    | p (SOM(x,t1,t2)) = 
          some x t1 (lower (push_bound te x dummytype) (i+1) t2 (Object+1))
    | p (TOP(k)) = 
          top(k)
    | p (RCD(fields)) = 
          rcd (map (fn (l,t) => (l, lower te i t Object)) fields)
  in wrap MUCHODEBUG "lower"
    (fn () => p (whn te t))
    (fn () => (write (makestring i); write " in "; prt te t))
    (fn t' => prt te t')
  end

val M = Id.intern "M"
val x = Id.intern "x"
val y = Id.intern "y"
val f = Id.intern "f"
val a = Id.intern "a"
val obj = Id.intern "obj"
val v = Id.intern "v"
val methods = Id.intern "methods"
val state = Id.intern "state"

fun build_id t = ABS(v, t, VAR(0))

(* Assumptions used in the following:
     t valid in te
     te = te' + MethodsM + M + _ + Rep + x
*)

val xOff = ~1
val RepOff = 0
val MOff = 2
val MethodsMOff = 3

fun pos te i t Object = let
  fun p (TVAR(i')) = 
      if (i = i') 
        then ABS(v, tvar(i+RepOff), 
               PACK(
                 RCDV[
                   (state, VAR(0)),
                   (methods, PROJ(VAR(i+xOff+1),methods))
                   ],
                 appt(tvar (Object+1)) (tvar (i+MOff+1)),
                 tvar(i+RepOff+1)))
        else build_id (tvar(i'))
    | p (ABST(n,k1,t2)) = 
        if Typ.occurs_free_in i t
           then raise BadObjDecl 
                 ("representation type appears in an operator")
           else build_id t
    | p (APPT(t1,t2)) = 
        if Typ.occurs_free_in i t
           then raise BadObjDecl 
                 ("representation type appears in an operator application")
           else build_id t
    | p (ARROW(t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_binding te' a dummytype
        in
          ABS(f, arrow t1 t2,
           ABS(a, lower te' (i+1) (relocate 1 t1) (Object+1),
            APP(pos te'' (i+2) (relocate 2 t2) (Object+2),
                APP(VAR(1),
                    APP(neg te'' (i+2) (relocate 2 t1) (Object+2), VAR(0))))))
        end
    | p (ALL(A,t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_bound te' A dummytype
        in
          if occurs_free_in i t1
            then raise BadObjDecl 
                 ("representation type appears in the bound of a quantifier")
            else ();
          ABS(f, all A t1 t2,
           TABS(A, relocate 1 t1,
            APP(pos te'' (i+2) (relocate_above 1 1 t2) (Object+2),
                TAPP(VAR(1), tvar(0)))))
        end
    | p (SOM(A,t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_binding (push_bound te' A dummytype) a dummytype
        val te''' = push_bound te'' A dummytype
        in
          if occurs_free_in i t1
            then raise BadObjDecl 
                 ("representation type appears in the bound of a quantifier")
            else ();
          ABS(f, some A t1 t2,
           OPEN(A, a, VAR(0),
             PACK(APP(pos te'' (i+3) (relocate 2 t2) (Object+3), VAR(0)),
                  some A (relocate 3 t1) 
                         (lift te''' (i+4) 
                               (relocate_above 1 3 t2)
                               (Object+4)),
                  tvar(1))))
        end
    | p (TOP(k)) = 
        build_id (top(k))
    | p (RCD(fields)) = 
        ABS(v, rcd(fields),
          RCDV (map (fn (l,t) => 
                        (l, APP(pos (push_binding te v dummytype) 
                                    (i+1) (relocate 1 t) (Object+1), 
                                PROJ(VAR(0),l)))) fields))
  in wrap MUCHODEBUG "pos"
    (fn () => p (whn te t))
    (fn () => (write (makestring i); write " in "; prt te t))
    (fn e => TrmHelper.prt te e)
  end

and neg te i t Object = let
  fun p (TVAR(i')) = 
      if (i = i') 
        then raise BadObjDecl ("representation type appears negatively")
        else build_id (tvar(i'))
    | p (ABST(n,k1,t2)) = 
        if Typ.occurs_free_in i t
           then raise BadObjDecl 
                 ("representation type appears in an operator")
           else build_id t
    | p (APPT(t1,t2)) = 
        if Typ.occurs_free_in i t
           then raise BadObjDecl 
                 ("representation type appears in an operator application")
           else build_id t
    | p (ARROW(t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_binding te' a dummytype
        in
          ABS(f, arrow (lift te i t1 Object) (lower te i t2 Object),
           ABS(a, relocate 1 t1,
            APP(neg te''(i+2) (relocate 2 t2) (Object+2),
                APP(VAR(1),
                    APP(pos te'' (i+2) (relocate 2 t1) (Object+2), VAR(0))))))
        end
    | p (ALL(A,t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_bound te' A dummytype
        in
          if occurs_free_in i t1
            then raise BadObjDecl 
                 ("representation type appears in the bound of a quantifier")
            else ();
          ABS(f, all A t1 (lower te i t2 Object),
           TABS(A, relocate 1 t1,
            APP(neg te'' (i+2) (relocate_above 1 1 t2) (Object+2),
                TAPP(VAR(1), tvar(0)))))
        end
    | p (SOM(A,t1,t2)) = let
        val te' = push_binding te f dummytype
        val te'' = push_binding (push_bound te' A dummytype) a dummytype
        val te''' = push_bound te'' A dummytype
        in
          if occurs_free_in i t1
            then raise BadObjDecl 
                 ("representation type appears in the bound of a quantifier")
            else ();
          ABS(f, some A t1 (lower te i t2 Object),
           OPEN(A, a, VAR(0),
             PACK(APP(neg te'' (i+3) (relocate 2 t2) (Object+3), VAR(0)),
                  some A (relocate 3 t1) 
                         (lower te''' (i+4) 
                                (relocate_above 1 3 t2)
                                (Object+4)),
                  tvar(1))))
        end
    | p (TOP(k)) = 
        build_id (top(k))
    | p (RCD(fields)) = 
        ABS(v, rcd(fields),
          RCDV (map (fn (l,t) => 
                        (l, APP(neg (push_binding te v dummytype) 
                                    (i+1) (relocate 1 t) (Object+1), 
                                PROJ(VAR(0),l)))) fields))
  in wrap MUCHODEBUG "neg"
    (fn () => p (whn te t))
    (fn () => (write (makestring i); write " in "; prt te t))
    (fn e => TrmHelper.prt te e)
  end

fun build_method te m t selfid = let
  val Object = TEnv.index te (Id.intern "Object")
  in
      TABS(M, tvar(0),
        ABS(obj, appt (tvar (Object-3)) (tvar 0),
          OPEN(selfid,x,
            VAR(0),
            APP(pos te 1 t Object, 
                APP(PROJ(PROJ(VAR(0),methods),m), PROJ(VAR(0), state))
            ))))
  end

fun build_objecttype name selfid method_decls te env = let
  val namestr = Id.tostr name
  val method_vector_name = Id.intern (namestr ^ "M")
  val te_with_self = push_bound te selfid dummytype
  val d_method_decls = 
      map (fn (m,pt) => (m, TypHelper.debruijnify te_with_self pt))
          method_decls
  val method_types = map (fn (m,t) => (m, arrow (tvar 0) t)) d_method_decls
  val method_vector_type = abst selfid star (rcd method_types)
  val te_with_mvt = push_abbrev te method_vector_name method_vector_type
  val env_with_mvt = push_env_null env method_vector_name

  val rel_method_decls = 
      map (fn(m,t)=> (m, Typ.relocate_above 2 3 (Typ.relocate 1 t))) 
          d_method_decls
  val te_with_theworks = 
      (push_binding 
        (push_bound 
          (push_binding 
            (push_bound te_with_mvt M dummytype)
            obj dummytype)
          selfid dummytype)
        x dummytype)
  val methods = map (fn(m,t) => (m, build_method te_with_theworks m t selfid)) 
                    rel_method_decls

  val renamed_methods = 
      map (fn (m,t) => (Id.intern (namestr ^ "'" ^ (Id.tostr m)), t))
           methods
  val typed_methods = 
      map (fn (m,e) => (m, e, Synth.synth te_with_mvt e)) renamed_methods

  fun build_te_and_env (te,env) [] = (te,env,0)
    | build_te_and_env (te,env) ((m,e,t)::rest) = let
      val (te',env',i) = build_te_and_env (te,env) rest
      val compiled = if !eval 
                        then Eval.eval te' env' m (Trm.relocate i e)
                        else Eval.dummycode
      in 
         (push_binding te' m (Typ.relocate i t), 
          push_env_binding env' m compiled,
          i+1)
      end
  val (te_with_methods, env_with_methods, _) = 
      build_te_and_env (te_with_mvt,env_with_mvt) typed_methods  

  val object_type = TypHelper.debruijnify te_with_methods
                      (PREAPPT(PRETVAR(Id.intern "Object"),
                               PRETVAR(method_vector_name)))
  val (final_te,final_env) = 
      (push_abbrev te_with_methods name object_type,
       push_env_null env_with_methods name)

  in
    write (Id.tostr method_vector_name);
    write " = ";
    TypHelper.prt_outer te method_vector_type;
    write "\n";

    write (Id.tostr name);
    write " = ";
    TypHelper.prt_outer te_with_methods object_type;
    write "\n";
    
    mapunit 
      (fn(m,e,t)=> (write (Id.tostr m);
                    setb();
                    if not (!DEBUG) then () else
                      (write " = ";
                       TrmHelper.prt te_with_mvt e;
                       ubreak 0);
                    write " : ";
                    TypHelper.prt_outer te_with_mvt t;
                    endb();
                    write "\n"))
      typed_methods;

    (final_te, final_env, 2+((*List required because of 0.90 bug?*)
                             List.length method_decls)) 
  end

(* ------------------------------------------------------------------------ *)

(* Translation:

        class cpoint : CPoint from point : Point =
          vars c1: Color = red,
               c2: Color = blue
          inherit setX,getX
          with
            setC = fun(c:Color) state@c1 = c
            getC = state@c1
          end

   becomes
   
        cpointClass = 
           extend [PointM] [CPointM] [{c1:Color,c2:Color}] pointClass
             {c1=red,c2=blue}
             (fun(FinalR<Top)
              fun(get: FinalR->{c1:Color,c2=Color})
              fun(put: FinalR->{c1:Color,c2=Color}->FinalR)
              fun(super: PointM(FinalR))
              fun(self: CPointM(FinalR))
                let putc1 = fun(f:FinalR) fun(x:Color)
                               let s = get f in
                                  put f {c1=x, c2=s.c2}
                               end
                in let 
                    putc2 = fun(f:FinalR) fun(x:Color)
                               let s = get f in
                                  put f {c1=s.c1, c2=x}
                               end
                in
                    {setC = fun(state:FinalR) fun(c:Color) putc1 state c,
                     getC = fun(state:FinalR) (get state).c1,
                     setX = super.setX,
                     getX = super.getX}
                end end)

        cpoint'new = new [PointM] cpointClass
*)

val extend = Id.intern "extend"
val FinalR = Id.intern "FinalR"
val get = Id.intern "get"
val put = Id.intern "put"
val super = Id.intern "super"
val self = Id.intern "self"
val state = Id.intern "state"
val Top = PRETOP(PRESTAR)
val new = Id.intern "new"
val f = Id.intern "f"
val x = Id.intern "x"
val s = Id.intern "s"

fun putname i = Id.intern("put" ^ (Id.tostr i))

fun build_classdecl name selftype supername supertype 
                    instancevars
                    inherits virtual method_bodies 
                    te env                              
  = let
  val namestr = Id.tostr name
  val supernamestr = Id.tostr supername
  val class_name = Id.intern (namestr ^ "Class")
  val superclass_name = Id.intern (supernamestr ^ "Class")
  val SelfM = Id.intern ((Id.tostr selftype) ^ "M")
  val SuperM = Id.intern ((Id.tostr supertype) ^ "M")
  val new_name = Id.intern (namestr ^ "'new")
 
  val statetype = PRERCD(map (fn (i,pt,_) => (i,pt)) instancevars)
  val initstate = PRERCDV(map (fn (i,_,ptm) => (i,ptm)) instancevars)

  val methodbuilderbody = 
        PRERCDV
          ((map (fn (l,m) => (l, PREABS(state, PRETVAR(FinalR), m)))
                method_bodies)

           @
           (map (fn (l) => (l, PREPROJ(PREVAR(super),l)))
                inherits)
           @
           (map (fn (l) => (l, PREPROJ(PREVAR(self),l)))
                virtual)
          )
  fun cpyfield i (i',_,_) = 
        (i', if (Id.== i i') then PREVAR(x) else PREPROJ(PREVAR(s), i'))
  fun lets_and_body [] = methodbuilderbody
    | lets_and_body ((i,pt,ptm)::rest) = 
        PRELET(putname(i), 
          PREABS(f,PRETVAR(FinalR),
            PREABS(x,pt,
              PRELET(s,
                PREAPP(PREVAR(get),PREVAR(f)),
                PREAPP(
                  PREAPP(PREVAR(put),PREVAR(f)),
                  PRERCDV(map (cpyfield i) instancevars))))),
          lets_and_body rest)

  val methodbuilder = 
        PRETABS(FinalR, Top,
        PREABS(get, PREARROW(PRETVAR(FinalR), statetype),
        PREABS(put, PREARROW(PRETVAR(FinalR), 
                             PREARROW(statetype, PRETVAR(FinalR))),
        PREABS(super, PREAPPT(PRETVAR(SuperM), PRETVAR(FinalR)),
        PREABS(self, PREAPPT(PRETVAR(SelfM), PRETVAR(FinalR)),
          lets_and_body instancevars
          )))))

  val pre_class_decl =
        PREAPP(
          PREAPP(
            PREAPP(
              PRETAPP(
                PRETAPP(
                  PRETAPP(PREVAR(extend), PRETVAR(SuperM)), 
                  PRETVAR(SelfM)),
                statetype),
              PREVAR(superclass_name)),
            initstate),
          methodbuilder)
  val class_decl = TrmHelper.debruijnify te pre_class_decl
  val class_decl_type = top star
  val class_decl_type = Synth.synth te class_decl 
  val compiled_class_decl = if !eval 
                               then Eval.eval te env class_name class_decl
                               else Eval.dummycode

  val te_with_class = push_binding te class_name class_decl_type
  val env_with_class = push_env_binding env class_name compiled_class_decl

  val pre_new_decl = 
        PREAPP(
          PRETAPP(PREVAR(new), PRETVAR(SelfM)),
          PREVAR(class_name))
  val new_decl = TrmHelper.debruijnify te_with_class pre_new_decl
  val new_decl_type = Synth.synth te_with_class new_decl 
  val compiled_new_decl = if !eval
                            then Eval.eval te_with_class env_with_class 
                                           new_name new_decl
                            else Eval.dummycode

  val te_with_new = push_binding te_with_class new_name new_decl_type
  val env_with_new = push_env_binding env_with_class new_name compiled_new_decl

  in 
     setb();
     write (Id.tostr class_name); 
     ubreak 2;
     if not (!DEBUG) then () else
       (write " = ";
        TrmHelper.prt te class_decl;
        ubreak 2);
     write " : ";
     TypHelper.prt_outer te class_decl_type;
     endb();
     write "\n";

     setb();
     write (Id.tostr new_name);
     ubreak 2;
     if not (!DEBUG) then () else
       (write " = ";
        TrmHelper.prt te_with_class new_decl;
        ubreak 2);
     write " : ";
     TypHelper.prt_outer te_with_class new_decl_type;
     endb();
     write "\n";

     (te_with_new, env_with_new, 2)
  end

end
