functor Synth(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    structure TrmHelper: TRMHELPER
	    sharing Typ.Globals = Globals
	        and Trm.Typ = Typ
	        and TypHelper.Typ = Typ
	        and TrmHelper.Trm = Trm
	        and TypHelper.TEnv = TEnv
	        and TypHelper.Typ = Typ
	        and Trm.TEnv = TEnv
	        and Typ.Id = Globals.Id
	    ) : SYNTH = struct

structure Globals = Globals
structure Typ = Typ
structure Trm = Trm
structure TEnv = TEnv
structure TypHelper = TypHelper
open Globals
open TypHelper
open TEnv
open Trm
open Typ

fun setquerytenv te = ()

exception IllTyped of unit->unit
exception SynthFailure of unit->unit

val DEBUG = createflag ["leq","Leq"] false
val DEBUGKSYNTH = createflag ["ksynth"] false
val DEBUGSYNTH = createflag ["synth"] false

(* This module is not implemented very efficiently in this version.
   For one thing, it must calculate the kinds of things from the context; 
   it would be more efficient to include kinding information in type 
   environments.  Also, kinds are repeatedly calculated during subtype
   testing (for example) to make sure that the kinds of corresponding
   subphrases are identical. *)

fun ksynth te t = let
  fun ks (TVAR(i)) = lookup_bound_or_abbrev te i
                          (fn teOFt => ksynth te teOFt)
                          (fn teOFt => ksynth te teOFt)
    | ks (ABST(a,k1,t2)) = let
        val k2 = ksynth (push_bound te a (top k1)) t2
        in karrow k1 k2
        end
    | ks (APPT(tf,ta)) = let
        val kf = ksynth te tf
        val ka = ksynth te ta
        in case kf of
             KARROW(k1,k2) => 
               if keq te k1 ka
             	  then k2
             	  else raise SynthFailure 
             	        (fn()=> (write "Cannot apply operator of kind ";
             	  	    TypHelper.kprt te kf; ubreak 3; write " to ";
             	  	    TypHelper.kprt te ka))
           | _ => raise SynthFailure 
             	        (fn()=> (write "Cannot apply kind "; 
             	  	    TypHelper.kprt te kf))
        end
    | ks (ARROW(t1,t2)) = 
        if (keq te star (ksynth te t1)) andalso (keq te star (ksynth te t2))
          then star
          else raise SynthFailure 
             	        (fn()=> (TypHelper.prt_outer te t; 
             	        ubreak 0; write " is ill formed"))
    | ks (ALL(a,t1,t2)) = 
        if (keq te star (ksynth (push_bound te a t1) t2)) 
                                (* $$$ was dummytype, not t1; bug? *)
             then star
             else raise SynthFailure
             	        (fn()=> (TypHelper.prt_outer te t; 
             	        ubreak 0; write " is ill formed"))
    | ks (SOM(a,t1,t2)) = 
        if (keq te star (ksynth (push_bound te a t1) t2))
                                (* $$$ was dummytype, not t1; bug? *)
             then star
             else raise SynthFailure 
             	        (fn()=> (TypHelper.prt_outer te t; 
             	        ubreak 0; write " is ill formed"))
    | ks (TOP(k)) = k
    | ks (RCD(l)) = if (forall (fn (_,t) => keq te star (ksynth te t)) l)
    			 then star
    			 else raise SynthFailure 
				     (fn()=> (TypHelper.prt_outer te t; 
				     ubreak 0; write " is ill formed"))
  in wrap DEBUGKSYNTH "ksynth"
    (fn () => ks t)
    (fn () => (prt te t; ubreak 0; write " in "; TypHelper.prt_tenv te))
    (fn k => kprt te k)
  end

fun samekind te s t = keq te (ksynth te s) (ksynth te t)

exception notWHNF of Typ.T

fun promote te t = let
  fun p (TVAR(i))     = lookup_bound_or_abbrev te i
  		  	      (fn teOFi => teOFi)
  			      (fn teOFi => raise notWHNF t)
    | p (APPT(tf,ta)) = appt (promote te tf) ta
    | p (TOP(_))      = raise SynthFailure
                                          (fn()=> (TypHelper.prt_outer te t; 
                                          write " "; ubreak 0; 
                                          write "cannot be promoted"))
    | p _                = t
  in p t
  end

fun basis te t = let
  fun b (TVAR(i))     = lookup_bound_or_abbrev te i
  			      (fn teOFi => basis te (whn te teOFi))
  			      (fn teOFi => raise notWHNF(t))
    | b (APPT(tf,ta)) = (case lambdabasis te tf of
    			        NONE => raise SynthFailure
                                          (fn()=> (TypHelper.prt_outer te t; 
                                          write " "; ubreak 0; 
                                          write "has no basis"))
    			      | SOME(a,tf1,tf2) => 
    			      	  basis te (whn te (tsubst_top ta tf2)))
    | b _                = t
  in b t
  end

and lambdabasis te t = 
  case (basis te t) of
      (ABST(n,t1,t2)) => SOME(n,t1,t2)
    | _                  => NONE

fun arrowbasis te t = 
  case (whn te (basis te t)) of
      (ARROW(t1,t2)) => (t1,t2)
    | _ => raise SynthFailure 
		   (fn()=> (TypHelper.prt_outer te t; write " ";
		   ubreak 0; write "has no arrow supertypes"))

fun allbasis te t = 
  case (whn te (basis te t)) of
      (ALL(n,t1,t2)) => (n,t1,t2)
    | _ => raise SynthFailure 
		   (fn()=> (TypHelper.prt_outer te t; write " ";
		   ubreak 0; write "has no quantifier supertypes"))

fun somebasis te t = 
  case (whn te (basis te t)) of
      (SOM(n,t1,t2)) => (n,t1,t2)
    | _ => raise SynthFailure 
		   (fn()=> (TypHelper.prt_outer te t; write " ";
		   ubreak 0; write "has no existential supertypes"))

fun rcdbasis te t = 
  case (whn te (basis te t)) of
      (RCD(fields)) => fields
    | _ => raise SynthFailure 
		   (fn()=> (TypHelper.prt_outer te t; write " ";
		   ubreak 0; write "has no record supertypes"))

(* Check whether a WHN type has the form "Top(K) T1..Tn" *)
fun eqv_to_top te (APPT(tf,ta)) = eqv_to_top te tf
  | eqv_to_top te (TOP(_))      = true
  | eqv_to_top te _             = false

(* ---------------------------------------------------------------------- *)
val leq_envs = createflag ["leqenvs"] false

(* Some variants of the quantifier subtyping rule. *)
val castagna = createflag ["castagna"] false
val equalbounds = createflag ["equalbounds"] false

fun leq te s t = let
  fun ok () = samekind te s t
  in wrap DEBUG "leq"
    (fn () => if ok() then leqOK te s t else false)
    (fn () => (prt te s; write " "; ubreak 0; write "<= "; prt te t;
               if not (!leq_envs) then () else
               (write "   "; ubreak 0; write "in ";
               prt_tenv te) ))
    (fn b => write (if b then "Yes" else "No"))
  end

and leqOK te s t = 
  wrap DEBUG "leq"
    (fn () => leqN te s (whn te t))
    (fn () => (prt te s; write " "; ubreak 0; write "<= "; prt te t;
               if not (!leq_envs) then () else
               (write "   "; ubreak 0; write "in ";
               prt_tenv te) ))
    (fn b => write (if b then "Yes" else "No"))

and leqN te s tN = let 
  fun l s tN = 
    if eqv_to_top te tN
      then true
      else NleqN te (whn te s) tN
  in wrap DEBUG "leqN"
    (fn () => l s tN)
    (fn () => (prt te s; write " "; ubreak 0; write "<=N "; prt te tN))
    (fn b => write (if b then "Yes" else "No"))
  end

and NleqN te sN tN = let
  fun l (TVAR(i)) tN =
        (case tN of 
            TVAR(i') => (i = i')
          | _ => false)
        orelse
        (leqN te (promote te sN) tN)
    | l (ARROW(s1,s2)) (ARROW(t1,t2)) =
        (leq te t1 s1)
        andalso
        (leq te s2 t2)
    | l (ALL(n,s1,s2)) (ALL(n',t1,t2)) =
          (leq te t1 s1)
        andalso
          (if !equalbounds then (leq te s1 t1) else true)
        andalso
          (if !castagna
            then (leq (push_bound te n (top(ksynth te t1))) s2 t2)
            else (leq (push_bound te n t1) s2 t2))
    | l (SOM(n,s1,s2)) (SOM(n',t1,t2)) =
        (leq te s1 t1)
        andalso
        (leq (push_bound te n s1) s2 t2)
    | l (ABST(n,ks1,s2)) (ABST(n',kt1,t2)) =
        (keq te ks1 kt1)
        andalso
        (leq (push_bound te n (top ks1)) s2 t2)
    | l (APPT(s1,s2)) tN =
        ((betaeq te sN tN) andalso (let val _ = ksynth te sN in true end))
        orelse
        (leqN te (promote te sN) tN)
    | l (RCD(ls)) (RCD(lt)) = 
        forall 
          (fn (i,t) => case Id.lookup_field i ls of
          		   NONE => false
          		 | SOME(s) => leq te s t)
          lt
    | l sx tx = false
  in wrap DEBUG "NleqN"
    (fn () => l sN tN)
    (fn () => (prt te sN; write " "; ubreak 0; write "N<=N "; prt te tN))
    (fn b => write (if b then "Yes" else "No"))
  end

fun eqv te s t = (leq te s t) andalso (leq te t s)

(* ---------------------------------------------------------------------- *)

(* $$$ belongs in Globals *)
fun curry f (x,y) = f x y

fun synth te e = let
  fun syn (VAR(i)) =
        lookup_binding te i
    | syn (ABS(x,t1,e)) = let
        val _ = case (ksynth te t1) of
                    STAR => ()
                  | _ => raise SynthFailure 
                           (fn()=> (TypHelper.prt_outer te t1; 
                           ubreak 0; write " does not have kind *"))
        val t2 = synth (push_binding te x t1) e
        in arrow t1 (relocate ~1 t2)
        end
    | syn (APP(f,v)) = let
        val tf = synth te f
        (* $$$ why form the whn here (and similarly below)? *)
        val (tf1,tf2) = arrowbasis te (whn te tf)
        val tv = synth te v
        in if (leq te tv tf1) 
             then tf2
             else raise SynthFailure
                     (fn()=> (write "Ill-typed application ";
                     ubreak 0; write "of operator type ";
                     setb();
                     TypHelper.prt_outer te tf; 
                     endb{}; write " ";
                     ubreak 0; write "to arg type ";
                     setb();
                     TypHelper.prt_outer te tv; 
                     endb()))
        end
    | syn (TABS(a,t1,e)) = let
        val _ = ksynth te t1
        val t2 = synth (push_bound te a t1) e
        in all a t1 t2
        end
    | syn (TAPP(e,s)) = let
        val t = synth te e
        val _ = ksynth te s
        val (n,t1,t2) = allbasis te (whn te t)
        in if (leq te s t1) 
             then tsubst_top s t2
             else raise SynthFailure 
                     (fn()=> (write "Ill-typed type application of ";
                     ubreak 0; write "operator type ";
                     setb();
                     TypHelper.prt_outer te t; 
                     endb(); write " ";
                     ubreak 0; write "to arg type ";
                     TypHelper.prt_outer te s))
        end
    | syn (RCDV(fields)) =
        rcd (map 
              (fn (i,e) => (i, synth te e))
              fields)
    | syn (PROJ(e,i)) = let
        val et = (synth te e)
        val fields = rcdbasis te (whn te et)
        in case (Id.lookup_field i fields) of
                 NONE => raise SynthFailure
                           (fn()=> (write "Ill-typed projection on field ";
                           write (Id.tostr i); write " "; ubreak 0;
                           write "of "; setb(); 
                           TypHelper.prt_outer te et; endb()))
               | SOME(t) => t
        end
    | syn (OPEN(reptype,rep,package,body)) = let
        val package_type = synth te package
        val (a,p1,p2) = somebasis te (whn te package_type)
        val body_te = (push_binding (push_bound te reptype p1) rep p2)
        val result_type = synth body_te body
        in if occurs_free_in 1 result_type
             then raise SynthFailure
                              (fn()=> (write "Ill-typed open: the body type ";
                              ubreak 5; setb(); 
                              TypHelper.prt_outer body_te result_type; 
                              write " "; endb(); ubreak 0; 
                              write "has the representation type variable "; 
                              write (Id.tostr reptype); write " free "))
             else relocate ~2 result_type
        end
    | syn (PACK(e,t,reptype)) = let
        val et = synth te e
        val (a,t1,t2) = somebasis te (whn te t)
        in 
           if (leq te (some a t1 t2) t)
             then ()
             else raise SynthFailure
                          (fn()=> (write "Ill-typed pack: the declared type ";
                          ubreak 5; setb();
                          TypHelper.prt_outer te t; write " ";
                          endb(); ubreak 0; 
                          write "is not equivalent to an existential type"));
           if (leq te et (Typ.tsubst_top reptype t2))
             then t   
             else raise SynthFailure
                              (fn()=> (write "Ill-typed pack: the body type ";
                              ubreak 5; setb();
                              TypHelper.prt_outer te et; write " ";
                              endb(); ubreak 0; 
                              write "doesn't have the declared type ";
                              ubreak 5; TypHelper.prt_outer te t))
        end
    | syn (ASSERT(e,t)) = let
        val et = synth te e
        in if (leq te et t)
             then t
             else raise SynthFailure
                              (fn()=> (write "Assertion failure: ";
                              write "synthesized type "; ubreak 5; setb();
                              TypHelper.prt_outer te et; write " "; endb();
                              ubreak 0; write "doesn't match declared type ";
                              ubreak 5;
                              TypHelper.prt_outer te t))
        end
    | syn (ABBREV(i,t,e)) = let
        val new_te = push_abbrev te i t
        val et = synth new_te e
        in Typ.tsubst_top t et 
        end
    | syn (LET(i,e1,e2)) = let
        val t1 = synth te e1
        val new_te = push_binding te i t1
        val t2 = synth new_te e2
        in relocate ~1 t2
        end
    | syn (PRIM(p,t)) = 
        t
  in (wrap DEBUGSYNTH "synth"
      (fn () => syn e)
      (fn () => (TrmHelper.prt_short te e))
      (fn t => TypHelper.prt_outer te t))
  handle SynthFailure(f) => 
    raise IllTyped (fn()=>(
             write "Typechecking error at ";
             ubreak 3; setb(); TrmHelper.prt te e; endb();
             ubreak 0; f()
             ))
  end

fun simplify te t = t

fun squash te t = t

end
